import tuxmake.__main__
import tuxmake.cli


def test_tuxmake_main_is_cli_main():
    assert tuxmake.__main__.main is tuxmake.cli.main


def test_calls_main_when___name___is___main__(monkeypatch, mocker):
    main = mocker.patch("tuxmake.__main__.main")
    monkeypatch.setattr(tuxmake.__main__, "__name__", "__main__")
    tuxmake.__main__.run()
    main.assert_called()



import argparse
import os
from pathlib import Path
import pytest
import sys
from tuxmake.build import BuildInfo
from tuxmake.cli import main as tuxmake
from tuxmake.exceptions import TuxMakeException


@pytest.fixture
def builds(home):
    return home / ".cache/tuxmake/builds"


@pytest.fixture(autouse=True)
def builder(mocker, linux, tmp_path):
    b = mocker.patch("tuxmake.cli.Build")
    b.return_value.passed = True
    b.return_value.failed = False
    b.return_value.source_tree = linux
    b.return_value.output_dir = tmp_path
    return b


@pytest.fixture
def build(builder):
    return builder.return_value


def args(called):
    return argparse.Namespace(**called.call_args[1])


class TestBuildRun:
    def test_basic_build(self, build, monkeypatch):
        monkeypatch.setattr(sys, "argv", ["tuxmake"])
        tuxmake()
        build.run.assert_called()


class TestTree:
    def test_basic_build(self, builder, monkeypatch):
        monkeypatch.setattr(sys, "argv", ["tuxmake"])
        tuxmake()
        assert args(builder).tree == "."

    def test_basic_build_with_directory(self, linux, builder):
        tree = str(linux)
        tuxmake("--directory", tree)
        assert args(builder).tree == tree

    def test_build_from_sys_argv(self, monkeypatch, builder):
        monkeypatch.setattr(sys, "argv", ["tuxmake", "--directory=/path/to/linux"])
        tuxmake()
        assert args(builder).tree == "/path/to/linux"

    def test_build_from_sys_argv_default_tree_is_cwd(self, monkeypatch, builder):
        monkeypatch.setattr(sys, "argv", ["tuxmake"])
        tuxmake()
        assert args(builder).tree == "."


class TestTargets:
    def test_config(self, builder):
        tuxmake("config")
        args(builder).targets == ["config"]

    def test_config_multiple(self, builder):
        tuxmake("config", "kernel")
        assert args(builder).targets == ["config", "kernel"]


class TestMakeVariables:
    def test_basic(self, builder):
        tuxmake("FOO=BAR")
        assert args(builder).make_variables == {"FOO": "BAR"}
        assert "targets" not in args(builder).__dict__

    def test_make_vars_and_targets(self, builder):
        tuxmake("FOO=BAR", "config")
        assert args(builder).make_variables == {"FOO": "BAR"}
        assert args(builder).targets == ["config"]

    def test_rejects_multiple_equal_signs(self, builder):
        with pytest.raises(SystemExit) as exit:
            tuxmake("FOO=BAR=QUX", "config")
        builder.assert_not_called()
        assert exit.value.code == 1


class TestKConfig:
    def test_kconfig(self, builder):
        tuxmake("--kconfig=olddefconfig")
        assert args(builder).kconfig == "olddefconfig"

    def test_kconfig_add(self, builder):
        tuxmake(
            "--kconfig-add=https://example.com/foo.config",
            "--kconfig-add=/path/to/local.config",
        )
        assert args(builder).kconfig_add == [
            "https://example.com/foo.config",
            "/path/to/local.config",
        ]


class TestToolchain:
    def test_toolchain(self, builder):
        tuxmake("--toolchain=gcc-10")
        assert args(builder).toolchain == "gcc-10"


class TestTargetArch:
    def test_target_arch(self, builder):
        tuxmake("--target-arch=arm64")
        assert args(builder).target_arch == "arm64"


class TestJobs:
    def test_jobs(self, builder):
        tuxmake("--jobs=300")
        assert args(builder).jobs == 300


class TestRuntime:
    def test_docker(self, builder):
        tuxmake("--runtime=docker")
        assert args(builder).runtime == "docker"


class TestImage:
    @pytest.fixture(autouse=True)
    def environment(self, monkeypatch):
        env = dict(os.environ)
        monkeypatch.setattr(os, "environ", env)
        return env

    def test_implies_runtime_docker(self, builder):
        tuxmake("--image=foobar")
        assert args(builder).runtime == "docker"

    def test_does_not_override_explicit_runtime(self, builder):
        tuxmake("--image=foobar", "--runtime=podman")
        assert args(builder).runtime == "podman"

    def test_not_passed_to_builder_class(self, builder):
        tuxmake("--image=foobar")
        assert "image" not in builder.call_args[1].keys()

    def test_sets_TUXMAKE_IMAGE(self, environment):
        tuxmake("--image=foobar")
        assert environment["TUXMAKE_IMAGE"] == "foobar"

    def test_backwards_compat_with_docker_image(self, environment, builder):
        tuxmake("--docker-image=foobar")
        assert environment["TUXMAKE_IMAGE"] == "foobar"
        assert "docker_image" not in builder.call_args[1].keys()


class TestImageRegistry:
    @pytest.fixture(autouse=True)
    def environment(self, monkeypatch):
        env = dict(os.environ)
        monkeypatch.setattr(os, "environ", env)
        return env

    def test_not_passed_to_builder_class(self, builder):
        tuxmake("--image-registry=localhost")
        assert "image_registry" not in builder.call_args[1].keys()

    def test_sets_env_variable(self, environment):
        tuxmake("--image-registry=localhost")
        assert environment["TUXMAKE_IMAGE_REGISTRY"] == "localhost"

    def test_implies_runtime_docker(self, builder):
        tuxmake("--image-registry=localhost")
        assert args(builder).runtime == "docker"

    def test_does_not_override_explicit_runtime(self, builder):
        tuxmake("--image-registry=localhost", "--runtime=podman")
        assert args(builder).runtime == "podman"


class TestFailFast:
    def test_fail_fast(self, builder):
        tuxmake("--fail-fast")
        assert args(builder).fail_fast


class TestVerbosity:
    def test_verbose(self, builder):
        tuxmake("--verbose")
        assert args(builder).verbose

    def test_quiet(self, builder):
        tuxmake("--quiet")
        assert args(builder).quiet

    def test_quiet_build(self, capsys):
        tuxmake("--quiet")
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""


class TestEnvironment:
    def test_environment(self, builder):
        tuxmake("--environment=FOO=BAR")
        assert args(builder).environment["FOO"] == "BAR"

    def test_multiple_environment_variables(self, builder):
        tuxmake("--environment=FOO=BAR", "--environment=BAZ=QUX")
        assert args(builder).environment["FOO"] == "BAR"
        assert args(builder).environment["BAZ"] == "QUX"

    def test_later_overrides_earlier(self, builder):
        tuxmake("--environment=FOO=BAR", "--environment=FOO=BAZ")
        assert args(builder).environment["FOO"] == "BAZ"

    def test_multiple_equal_signs(self, builder):
        tuxmake("--environment=OPTIONS=x=y")
        assert args(builder).environment["OPTIONS"] == "x=y"


class TestWrapper:
    def test_ccache(self, builder):
        tuxmake("--wrapper=ccache")
        assert args(builder).wrapper == "ccache"


class TestExceptions:
    def test_basic(self, builder, capsys):
        builder.side_effect = TuxMakeException("hello")
        with pytest.raises(SystemExit) as exit:
            tuxmake("-C", "/path/to/linux")
        assert exit.value.code == 1
        _, err = capsys.readouterr()
        assert "E: hello" in err


class TestBuildStatus:
    def test_exits_2_on_build_failure(self, builder, capsys):
        builder.return_value.failed = True
        builder.return_value.status = {
            "config": BuildInfo("PASS", 1),
            "kernel": BuildInfo("FAIL", 2),
        }
        with pytest.raises(SystemExit) as exit:
            tuxmake("-C", "/path/to/linux")
        assert exit.value.code == 2

        _, err = capsys.readouterr()
        assert "config: PASS" in err
        assert "kernel: FAIL" in err


class TestList:
    def test_list_architectures(self, builder, capsys):
        tuxmake("--list-architectures")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert "x86_64" in out
        assert "arm64" in out

    def test_list_toolchains(self, builder, capsys):
        tuxmake("--list-toolchains")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert "gcc" in out
        assert "clang" in out

    def test_list_docker_toolchains(self, builder, capsys):
        tuxmake("--runtime=docker", "--list-toolchains")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert "gcc" in out
        assert "gcc-10" in out
        assert "clang" in out
        assert "clang-10" in out

    def test_list_runtimes(self, builder, capsys):
        tuxmake("--list-runtimes")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert "null" in out
        assert "docker" in out


class TestPrintSupportMatrix:
    def test_print_support_matrix(self, builder, capsys):
        tuxmake("--print-support-matrix")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert out

    def test_print_support_matrix_docker(self, builder, capsys):
        tuxmake("--runtime=docker", "--print-support-matrix")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert "yes" in out
        assert "no" in out

    def test_output_colors_when_requested(self, builder, capsys):
        tuxmake("--runtime=docker", "--color=always", "--print-support-matrix")
        builder.assert_not_called()
        out, _ = capsys.readouterr()
        assert "\033" in out


class TestDebug:
    def test_shell(self, builder, mocker):
        run_cmd = builder.return_value.run_cmd
        tuxmake("--shell")
        run_cmd.assert_called_with(["bash"], interactive=True)

    def test_debug(self, builder, mocker):
        tuxmake("--debug")
        assert args(builder).debug


class TestOutputDir:
    def test_output_dir(self, builder):
        tuxmake("--output-dir=/path/to/output")
        assert args(builder).output_dir == Path("/path/to/output")

    def test_output_dir_relative_to_absolute(self, builder):
        cwd = Path.cwd()
        tuxmake("--output-dir=output")
        assert args(builder).output_dir == (cwd / "output")


class TestBuildDir:
    def test_build_dir(self, builder):
        tuxmake("--build-dir=/path/to/build")
        assert args(builder).build_dir == Path("/path/to/build")


class TestOptionsFromEnvironment:
    def test_options_from_environment(self, builder, monkeypatch):
        monkeypatch.setenv("TUXMAKE", "--debug --verbose")
        tuxmake("config")
        assert args(builder).debug
        assert args(builder).verbose

    def test_command_line_overrides_environment(self, builder, monkeypatch):
        monkeypatch.setenv("TUXMAKE", "--toolchain=gcc")
        tuxmake("--toolchain=clang")
        assert args(builder).toolchain == "clang"

    def test_command_line_plus_sys_argv(self, builder, monkeypatch):
        monkeypatch.setenv("TUXMAKE", "--toolchain=gcc")
        monkeypatch.setattr(sys, "argv", ["tuxmake", "--toolchain=clang"])
        tuxmake()
        assert args(builder).toolchain == "clang"


class TestConfigurationFiles:
    @pytest.fixture(autouse=True)
    def config_dir(self, home):
        confdir = home / ".config" / "tuxmake"
        confdir.mkdir(parents=True)
        return confdir

    def test_basics(self, builder, config_dir):
        with (config_dir / "ccache").open("w") as f:
            f.write("--wrapper=ccache\n")
        tuxmake("@ccache")
        assert args(builder).wrapper == "ccache"

    def test_default(self, builder, config_dir, monkeypatch):
        with (config_dir / "default").open("w") as f:
            f.write("--runtime=podman\n")
        monkeypatch.setattr(sys, "argv", [])
        tuxmake()
        assert args(builder).runtime == "podman"

    def test_escaping(self, builder, config_dir):
        with (config_dir / "escaping").open("w") as f:
            f.write("--environment=FOO='bar baz'\n")
        tuxmake("@escaping")
        assert args(builder).environment["FOO"] == "bar baz"

    def test_order(self, builder, config_dir):
        with (config_dir / "before").open("w") as f:
            f.write("--runtime=podman\n")
        with (config_dir / "after").open("w") as f:
            f.write("--kconfig=defconfig\n")
        tuxmake("@before", "--runtime=null", "--kconfig=tinyconfig", "@after")
        assert args(builder).runtime == "null"
        assert args(builder).kconfig == "defconfig"

    def test_comments(self, builder, config_dir):
        with (config_dir / "comments").open("w") as f:
            f.write("# this is a comment\n")
        tuxmake("@comments")
        assert "targets" not in args(builder)

    def test_missing_config_file(self, capsys):
        with pytest.raises(SystemExit) as exit:
            tuxmake("@missing")
        _, err = capsys.readouterr()

        assert exit.value.code == 1
        assert "E: missing configuration file: " in err

    def test_absolute_path(self, builder, home):
        with (home / "file").open("w") as f:
            f.write("--environment=FOO=BAR\n")
        tuxmake(f"@{home}/file")
        assert args(builder).environment["FOO"] == "BAR"

    def test_XDG_CONFIG_HOME(self, builder, home, monkeypatch):
        custom_config_dir = home / "customconfig"
        (custom_config_dir / "tuxmake").mkdir(parents=True)
        with (custom_config_dir / "tuxmake" / "myconfig").open("w") as f:
            f.write("--environment=CONFIG=custom")
        monkeypatch.setenv("XDG_CONFIG_HOME", str(custom_config_dir))
        tuxmake("@myconfig")
        assert args(builder).environment["CONFIG"] == "custom"


class TestKernel:
    def test_kernel(self, builder):
        tuxmake("--kernel-image=Image.bz2")
        assert args(builder).kernel_image == "Image.bz2"


@pytest.fixture
def check_call(mocker):
    return mocker.patch("subprocess.check_call")


class TestHooks:
    def test_pre_build_hook(self, check_call, linux):
        tuxmake('--pre-build-hook=echo "hello world"')
        check_call.assert_called_with(
            ["sh", "-c", 'echo "hello world"'], cwd=str(linux)
        )

    def test_failed_pre_build_hook_stops_build(self, build, linux):
        with pytest.raises(SystemExit):
            tuxmake("--pre-build-hook=false")
        build.run.assert_not_called()

    def test_post_build_hook(self, check_call, linux):
        tuxmake("--post-build-hook=git clean -dxf")
        check_call.assert_called_with(["sh", "-c", "git clean -dxf"], cwd=str(linux))

    def test_post_build_hook_failure_means_exit_failure(self):
        with pytest.raises(SystemExit) as exit:
            tuxmake("--post-build-hook=false")
        assert exit.value.code == 2

    def test_post_build__hook_does_not_run_if_build_failed(self, build, check_call):
        build.passed = False
        build.failed = True
        with pytest.raises(SystemExit) as exit:
            tuxmake("--post-build-hook=true")
        assert exit.value.code == 2
        check_call.assert_not_called()

    def test_results_hook(self, check_call, tmp_path):
        tuxmake("--results-hook=ls -1")
        check_call.assert_called_with(["sh", "-c", "ls -1"], cwd=str(tmp_path))

    def test_results_hook_failure_means_exit_failure(self):
        with pytest.raises(SystemExit) as exit:
            tuxmake("--results-hook=false")
        assert exit.value.code == 2

    def test_results_hook_does_not_run_if_build_failed(self, build, check_call):
        build.passed = False
        build.failed = True
        with pytest.raises(SystemExit):
            tuxmake("--results-hook=true")
        check_call.assert_not_called()


class TestCheckEnvironment:
    def test_check_environment(self, build):
        tuxmake("--check-environment")
        build.check_environment.assert_called()
        build.run.assert_not_called()



import re
import subprocess
import pytest

from tuxmake.build import Build
from tuxmake.exceptions import InvalidRuntimeError
from tuxmake.exceptions import RuntimePreparationFailed
from tuxmake.runtime import Runtime
from tuxmake.runtime import NullRuntime
from tuxmake.runtime import DockerRuntime
from tuxmake.runtime import DockerLocalRuntime
from tuxmake.runtime import PodmanRuntime
from tuxmake.runtime import PodmanLocalRuntime
from tuxmake.runtime import Terminated


@pytest.fixture
def build(linux):
    b = Build(linux)
    return b


class TestTerminated:
    def test_basics(self):
        with pytest.raises(Terminated):
            Terminated.handle_signal(9, "BOOM")


class TestGetRuntime:
    def test_null_runtime(self):
        assert isinstance(Runtime.get(None), NullRuntime)

    def test_docker_runtime(self):
        assert isinstance(Runtime.get("docker"), DockerRuntime)

    def test_docker_local_runtime(self):
        assert isinstance(Runtime.get("docker-local"), DockerLocalRuntime)

    def test_invalid_runtime(self):
        with pytest.raises(InvalidRuntimeError) as exc:
            Runtime.get("invalid")
        assert str(exc.value) == "Invalid runtime: invalid"
        with pytest.raises(InvalidRuntimeError):
            Runtime.get("xyz")


class TestRuntime:
    def test_invalid_runtime(self, monkeypatch):
        monkeypatch.setattr(Runtime, "name", "invalid")
        with pytest.raises(InvalidRuntimeError):
            Runtime()

    def test_run_cmd_interactive(self, Popen, mocker):
        get_command_line = mocker.patch(
            "tuxmake.runtime.Runtime.get_command_line", return_value=["/bin/bash"]
        )
        runtime = NullRuntime()
        runtime.run_cmd(["/bin/bash"], interactive=True)
        kwargs = Popen.call_args[1]
        get_command_line.assert_called_with(
            ["/bin/bash"], interactive=True, offline=True
        )
        assert kwargs["stdin"] is None
        assert kwargs["stdout"] is None
        assert kwargs["stderr"] is None

    def test_create_output_directory_on_prepare(self, tmp_path):
        runtime = NullRuntime()
        runtime.output_dir = tmp_path / "build" / "1"
        runtime.prepare()
        assert runtime.output_dir.exists()

    def test_create_output_directory_on_prepare_already_exists(self, tmp_path):
        runtime = NullRuntime()
        d = tmp_path / "build" / "1"
        d.mkdir(parents=True)
        runtime.output_dir = d
        runtime.prepare()
        assert runtime.output_dir.exists()


class TestNullRuntime:
    def test_get_command_line(self, build):
        assert NullRuntime().get_command_line(
            ["date"], interactive=False, offline=False
        ) == ["date"]

    def test_toolchains(self):
        runtime = NullRuntime()
        assert "gcc" in runtime.toolchains


@pytest.fixture
def container_id():
    return "0123456789abcdef"


class FakeGetImage:
    @pytest.fixture(autouse=True)
    def get_image(self, mocker):
        return mocker.patch("tuxmake.runtime.Runtime.get_image")


class TestContainerRuntime(FakeGetImage):
    @pytest.fixture(autouse=True)
    def spawn_container(self, mocker, container_id):
        return mocker.patch(
            "tuxmake.runtime.ContainerRuntime.spawn_container",
            return_value=container_id,
        )

    @pytest.fixture(autouse=True)
    def offline_available(self, mocker):
        return mocker.patch(
            "tuxmake.runtime.Runtime.offline_available", return_value=False
        )


class TestGetImage:
    def test_no_image(self):
        runtime = NullRuntime()
        with pytest.raises(Exception):
            runtime.get_image()

    def test_with_image(self):
        runtime = NullRuntime()
        runtime.set_image("myimage")
        assert runtime.get_image() == "myimage"


class TestDockerRuntime(TestContainerRuntime):
    def test_get_metadata(self, get_image, mocker):
        get_image.return_value = "tuxmake/theimage"
        mocker.patch(
            "subprocess.check_output",
            return_value=b"tuxmake/theimage@sha256:deadbeef",
        )
        metadata = DockerRuntime().get_metadata()
        assert metadata["image_name"] == "tuxmake/theimage"
        assert metadata["image_digest"] == "tuxmake/theimage@sha256:deadbeef"

    def test_prepare(self, get_image, mocker):
        get_image.return_value = "myimage"
        check_call = mocker.patch("subprocess.check_call")
        DockerRuntime().prepare()
        check_call.assert_called_with(["docker", "pull", "myimage"])

    def test_prepare_pull_only_once_a_day(self, get_image, mocker):
        get_image.return_value = "myimage"
        check_call = mocker.patch("subprocess.check_call")
        now = 1614000983
        mocker.patch("time.time", return_value=now)
        two_hours_ago = now - (2 * 60 * 60)
        two_days_ago = now - (2 * 24 * 60 * 60)
        mocker.patch(
            "tuxmake.cache.get", side_effect=(None, two_hours_ago, two_days_ago)
        )

        # first call
        PodmanRuntime().prepare()
        assert len(check_call.call_args_list) == 1

        # after 2 hours, no need to pull
        PodmanRuntime().prepare()
        assert len(check_call.call_args_list) == 1

        # after 2 days, pull again
        PodmanRuntime().prepare()
        assert len(check_call.call_args_list) == 2

    def test_start_container(self, container_id):
        runtime = DockerRuntime()
        runtime.start_container()
        assert runtime.container_id == container_id

    def test_cleanup(self, container_id, mocker):
        call = mocker.patch("subprocess.call")
        runtime = DockerRuntime()
        runtime.start_container()
        runtime.cleanup()
        cmd = call.call_args[0][0]
        assert cmd[0:2] == ["docker", "stop"]
        assert cmd[-1] == container_id

    def test_cleanup_before_container_exists(self):
        runtime = DockerRuntime()
        assert runtime.container_id is None
        runtime.cleanup()  # if this doesn't crash we are good

    def test_get_command_line(self):
        cmd = DockerRuntime().get_command_line(["date"], False)
        assert cmd[0:2] == ["docker", "exec"]
        assert cmd[-1] == "date"

    def test_environment(self, linux, spawn_container):
        runtime = DockerRuntime()
        runtime.environment["FOO"] = "BAR"
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--env=FOO=BAR" in cmd

    def test_caps(self, linux, spawn_container):
        runtime = DockerRuntime()
        runtime.caps = ["SYS_PTRACE"]
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--cap-add=SYS_PTRACE" in cmd

    def test_output_dir(self, linux, tmp_path, spawn_container):
        runtime = DockerRuntime()
        runtime.output_dir = tmp_path
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert f"--volume={tmp_path}:{tmp_path}" in cmd

    def test_source_dir(self, linux, tmp_path, spawn_container):
        runtime = DockerRuntime()
        runtime.source_dir = tmp_path
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert f"--volume={tmp_path}:{tmp_path}" in cmd

    def test_volume(self, linux, spawn_container):
        runtime = DockerRuntime()
        path = "/path/to/something"
        runtime.add_volume(path)
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert f"--volume={path}:{path}" in cmd

    def test_TUXMAKE_DOCKER_RUN(self, monkeypatch, spawn_container):
        monkeypatch.setenv(
            "TUXMAKE_DOCKER_RUN", "--hostname=foobar --env=FOO='bar baz'"
        )
        DockerRuntime().start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--hostname=foobar" in cmd
        assert "--env=FOO=bar baz" in cmd

    def test_set_user(self, spawn_container, mocker):
        mocker.patch("os.getuid", return_value=6666)
        check_call = mocker.patch("subprocess.check_call")
        runtime = DockerRuntime()
        runtime.set_user("tuxmake")
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--user=tuxmake" in cmd
        usermod = check_call.call_args[0][0]
        assert usermod[-4:] == ["usermod", "-u", "6666", "tuxmake"]

    def test_set_user_set_group(self, spawn_container, mocker):
        mocker.patch("os.getgid", return_value=7777)
        check_call = mocker.patch("subprocess.check_call")
        runtime = DockerRuntime()
        runtime.set_user("tuxmake")
        runtime.set_group("tuxmake")
        runtime.start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--user=tuxmake:tuxmake" in cmd
        groupmod = check_call.call_args[0][0]
        assert groupmod[-4:] == ["groupmod", "-g", "7777", "tuxmake"]

    def test_interactive(self):
        cmd = DockerRuntime().get_command_line(["bash"], True)
        assert "--interactive" in cmd
        assert "--tty" in cmd

    def test_bases(self):
        assert [
            t.name
            for t in DockerRuntime().base_images
            if not t.name.startswith("base-debian")
        ] == []

    def test_toolchain_images(self):
        images = [t.name for t in DockerRuntime().toolchain_images]
        assert "gcc" in images
        assert "clang" in images

    def test_toolchains(self):
        toolchains = DockerRuntime().toolchains
        assert "gcc" in toolchains
        assert "clang" in toolchains
        assert "llvm" in toolchains

    def test_listed_as_supported(self):
        assert "docker" in Runtime.supported()

    def test_str(self):
        assert str(DockerRuntime()) == "docker"


class TestDockerRuntimeSpawnContainer(FakeGetImage):
    def test_spawn_container(self, mocker, container_id):
        check_output = mocker.patch(
            "subprocess.check_output", return_value=container_id.encode("utf-8")
        )
        runtime = DockerRuntime()
        runtime.start_container()
        cmd = check_output.call_args[0][0]
        assert cmd[0:2] == ["docker", "run"]
        assert runtime.container_id == container_id


class TestDockerRuntimeOfflineAvailable(FakeGetImage):
    @pytest.fixture
    def runtime(self, container_id, mocker):
        mocker.patch(
            "tuxmake.runtime.DockerRuntime.spawn_container", return_value=container_id
        )
        mocker.patch("tuxmake.runtime.ContainerRuntime.prepare_image")
        r = DockerRuntime()
        r.prepare()
        return r

    def test_offline_available(self, runtime, mocker):
        mocker.patch("subprocess.check_output")
        assert runtime.offline_available

    def test_offline_not_available(self, runtime, mocker, capsys):
        mocker.patch(
            "subprocess.check_output",
            side_effect=subprocess.CalledProcessError(
                1, ["true"], output=b"some error"
            ),
        )
        assert not runtime.offline_available
        _, stderr = capsys.readouterr()
        assert re.match("W:.*(some error)", stderr)


class TestDockerLocalRuntime(TestContainerRuntime):
    def test_prepare_checks_local_image(self, get_image, mocker):
        get_image.return_value = "mylocalimage"
        check_call = mocker.patch("subprocess.check_call")
        runtime = DockerLocalRuntime()

        runtime.prepare()
        check_call.assert_called_with(
            ["docker", "image", "inspect", "mylocalimage"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

    def test_prepare_image_not_found(self, get_image, mocker):
        get_image.return_value = "foobar"
        mocker.patch(
            "subprocess.check_call",
            side_effect=subprocess.CalledProcessError(
                1, ["foo"], stderr="Image not found"
            ),
        )
        with pytest.raises(RuntimePreparationFailed) as exc:
            DockerLocalRuntime().prepare()
        assert "image foobar not found locally" in str(exc)

    def test_listed_as_supported(self):
        assert "docker-local" in Runtime.supported()

    def test_str(self):
        assert str(DockerLocalRuntime()) == "docker-local"


class TestPodmanRuntime(TestContainerRuntime):
    def test_prepare(self, get_image, mocker):
        get_image.return_value = "myimage"
        check_call = mocker.patch("subprocess.check_call")
        PodmanRuntime().prepare()
        check_call.assert_called_with(["podman", "pull", "myimage"])

    def test_get_command_line(self):
        cmd = PodmanRuntime().get_command_line(["date"], False)
        assert cmd[0:2] == ["podman", "exec"]
        assert cmd[-1] == "date"

    def test_listed_as_supported(self):
        assert "podman" in Runtime.supported()

    def test_no_user_option(self, get_image, spawn_container):
        PodmanRuntime().start_container()
        cmd = spawn_container.call_args[0][0]
        assert len([c for c in cmd if "--user=" in c]) == 0

    def test_str(self):
        assert str(PodmanRuntime()) == "podman"

    def test_TUXMAKE_PODMAN_RUN(self, monkeypatch, spawn_container):
        monkeypatch.setenv(
            "TUXMAKE_PODMAN_RUN", "--hostname=foobar --env=FOO='bar baz'"
        )
        PodmanRuntime().start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--hostname=foobar" in cmd
        assert "--env=FOO=bar baz" in cmd

    def test_selinux_label(self, get_image, spawn_container):
        PodmanRuntime().start_container()
        cmd = spawn_container.call_args[0][0]
        volumes = [o for o in cmd if o.startswith("--volume=")]
        assert all([v.endswith(":z") for v in volumes])

    def test_logging_level(self, spawn_container):
        PodmanRuntime().start_container()
        cmd = spawn_container.call_args[0][0]
        assert "--log-level=ERROR" in cmd


class TestPodmanLocalRuntime(TestContainerRuntime):
    def test_prepare_checks_local_image(self, get_image, mocker):
        get_image.return_value = "mylocalimage"
        check_call = mocker.patch("subprocess.check_call")
        runtime = PodmanLocalRuntime()

        runtime.prepare()
        check_call.assert_called_with(
            ["podman", "image", "inspect", "mylocalimage"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

    def test_prepare_image_not_found(self, get_image, mocker):
        get_image.return_value = "foobar"
        mocker.patch(
            "subprocess.check_call",
            side_effect=subprocess.CalledProcessError(
                1, ["foo"], stderr="Image not found"
            ),
        )
        with pytest.raises(RuntimePreparationFailed) as exc:
            PodmanLocalRuntime().prepare()
        assert "image foobar not found locally" in str(exc)

    def test_listed_as_supported(self):
        assert "podman-local" in Runtime.supported()

    def test_str(self):
        assert str(PodmanLocalRuntime()) == "podman-local"



from tuxmake import deprecated


class TestGetenv:
    def test_warning(self, monkeypatch, capsys):
        monkeypatch.setenv("FOO", "something")
        deprecated.getenv("FOO", "BAR")
        _, err = capsys.readouterr()
        assert "FOO is deprecated" in err
        assert "use BAR instead" in err



import pytest

import tuxmake.exceptions
from tuxmake.arch import Native
from tuxmake.target import Compression, Target, Config


@pytest.fixture
def build(mocker):
    b = mocker.MagicMock()
    b.target_arch = Native()
    b.target_overrides = {"kernel": b.target_arch.artifacts["kernel"]}
    b.kconfig = ["defconfig"]
    return b


@pytest.fixture
def config(build):
    return Config("config", build)


def test_unsupported(build):
    with pytest.raises(tuxmake.exceptions.UnsupportedTarget):
        Target("foobarbaz", build)


def test_comparison(build):
    t1 = Target("kernel", build)
    t2 = Target("kernel", build)
    assert t1 == t2
    assert t1 in [t2]


class TestInstantiateAll:
    @pytest.mark.parametrize("target_name", Target.supported())
    def test_instantiate(self, build, target_name):
        target = Target(target_name, build)
        assert type(target.commands) is list
        assert type(target.artifacts) is dict


class TestConfig:
    def test_name(self, config):
        assert config.name == "config"

    def test___str__(self, config):
        assert str(config) == "config"

    def test_description(self, config):
        assert isinstance(config.description, str)

    def test_artifacts(self, config):
        assert config.artifacts["config"] == ".config"

    def test_does_nothing_if_dot_config_already_exists(self, config, build):
        build.kconfig = "defconfig"
        (build.build_dir / ".config").touch()
        config.prepare()
        assert config.commands == []


class TestDebugKernel:
    def test_commands(self, build):
        debugkernel = Target("debugkernel", build)
        assert debugkernel.commands[0][0] == "{z}"
        assert debugkernel.commands[0][-1] == "{build_dir}/vmlinux"


class TestKernel:
    def test_gets_kernel_name_from_arch(self, build):
        kernel = Target("kernel", build)
        assert kernel.artifacts

    def test_depends_on_default(self, build):
        kernel = Target("kernel", build)
        assert kernel.dependencies == ["default"]


class TestModules:
    @pytest.fixture
    def modules(self, build):
        return Target("modules", build)

    def test_install_modules(self, modules):
        assert modules.commands[1][0:2] == ["{make}", "modules_install"]

    def test_strip_modules(self, modules):
        assert modules.makevars["INSTALL_MOD_STRIP"] == "1"

    def test_depends_on_config(self, modules):
        assert modules.dependencies == ["config"]


class TestDtbs:
    def test_commands(self, build):
        dtbs = Target("dtbs", build)
        assert dtbs.commands[0] == ["{make}", "dtbs"]
        assert dtbs.commands[3][1] == "dtbs_install"
        assert dtbs.makevars["INSTALL_DTBS_PATH"] == "{build_dir}/dtbsinstall/dtbs"

    def test_depends_on_config(self, build):
        dtbs = Target("dtbs", build)
        assert dtbs.dependencies == ["config"]

    def test_artifacts(self, build):
        dtbs = Target("dtbs", build)
        assert dtbs.artifacts["dtbs.tar.xz"] == "dtbs.tar.xz"


class TestDefault:
    def test_command(self, build):
        default = Target("default", build)
        assert default.commands == [["{make}"]]

    def test_depends_on_config(self, build):
        default = Target("default", build)
        assert default.dependencies == ["config"]


class TestTargzPkg:
    def test_wildcards_in_artifacts(self, build, tmp_path):
        filename = "linux-5.13.0-rc2+-x86.tar.gz"
        (tmp_path / filename).touch()

        targzpkg = Target("targz-pkg", build)
        artifacts = targzpkg.find_artifacts(tmp_path)
        assert artifacts[0][0] == filename
        assert artifacts[0][1].name == filename


class TestCompression:
    def test_invalid_compression(self):
        with pytest.raises(tuxmake.exceptions.UnsupportedCompression):
            Compression("unexisting")



import os
from pathlib import Path


def _resolve(var, default):
    home = os.getenv(var)
    if home:
        return Path(home)
    else:
        return Path.home() / default


def cache_dir():
    return _resolve("XDG_CACHE_HOME", ".cache") / "tuxmake"


def config_dir():
    return _resolve("XDG_CONFIG_HOME", ".config") / "tuxmake"
